import datetime
import collections
from lxml import etree

def multipay(config):
    root = etree.Element('multipay')

    prime_mandatory = ['project_id', 'amount', 'currency_code',
                       'success_url', 'abort_url']
    prime_optional = ['interface_version', 'language_code', 'timeout',
                      'email_customer', 'phone_customer', 'timeout_url',
                      'success_link_redirect']
    prime_lists = {
        'user_variables': 'user_variable',
        'reasons': 'reason'
    }

    notification_addresses = {
        'notification_urls': 'notification_url',
        'notification_emails': 'notification_email'
    }

    for attr_name in prime_mandatory:
        etree.SubElement(root, attr_name).text = str(getattr(config, attr_name))

    for attr_name in prime_optional:
        value = getattr(config, attr_name, None)

        if value is None:
            continue

        etree.SubElement(root, attr_name).text = __serialize(value)

    for collection_name, item_name in prime_lists.items():
        if not config.has(collection_name):
            continue

        collection = etree.SubElement(root, collection_name)
        for value in getattr(config, collection_name):
            etree.SubElement(collection, item_name).text = value

    for collection_name, item_name in notification_addresses.items():
        if not config.has(collection_name):
            continue

        collection = etree.SubElement(root, collection_name)
        addresses = __compact_notification_addresses(getattr(config, collection_name))
        for notify_on, address in addresses.items():
            node = etree.SubElement(collection, item_name)
            node.text = address
            if notify_on != 'default':
                node.set('notify_on', notify_on)

    etree.SubElement(root, 'su')

    return etree.tostring(root)


def transaction_request_by_params(params):
    root = etree.Element('transaction_request')
    root.set('version', '2')
    for name, value in params.iteritems():
        if name == 'transaction':
            for transaction_id in value:
                etree.SubElement(root, 'transaction').text = transaction_id
        else:
            etree.SubElement(root, name).text = __serialize(value)
    return etree.tostring(root)


def __compact_notification_addresses(addresses_):
    """
    Input::

        {
            'default': 'http://domain/notice',
            'event1': 'http://domain/notice',
            'event2': 'http://domain/notice',
            'event3': 'http://domain/notice3',
            'event4': 'http://domain/notice3'
        }

    Output::

        {
            'default': 'http://domain/notice',
            'event1,event2': 'http://domain/notice',
            'event3,event4': 'http://domain/notice3'
        }

    """
    result = {}
    addresses = dict(addresses_)
    default = addresses.pop('default', None)

    for key, value in __reverse_group_dict(addresses).items():
        result[','.join(value)] = key

    if default:
        result['default'] = default

    return result


def __reverse_group_dict(dictionary):
    result = collections.defaultdict(list)

    for key, value in dictionary.items():
        result[value].append(key)

    return result


def __serialize(value):
    if isinstance(value, bool):
        value = int(value)

    if isinstance(value, datetime.datetime):
        value = value.isoformat()

    if isinstance(value, int):
        value = str(value)

    return value



# -*- coding: utf-8 -*-

import re
import copy


class Config(object):
    def __init__(self, **params):
        self.update(params)

    def has(self, key):
        return key in self.__dict__.keys()

    def update(self, dict_):
        self.__dict__.update(dict_)
        return self

    def clone(self):
        return Config(**copy.deepcopy(self.__dict__))


def strip_reason(reason):
    """
    only the following characters are allowed:
    '0-9', 'a-z', 'A-Z', ' ', '+', ',', '-', '.'.
    Umlauts are replaced, e.g. ä -> ae.

    @see: https://www.sofort.com/integrationCenter-eng-DE/content/view/full/2513#h5-1
    """
    return re.sub('(?u)[^\w\ \+\-\.\,]', '', reason)


def as_list(value):
    if not isinstance(value, list):
        value = [value]
    return value



import xmltodict
from warnings import warn

import iso8601

from schematics.models import Model
from schematics.types import (URLType, StringType, IntType, BooleanType,
                              DateTimeType, DecimalType, EmailType)
from schematics.types.compound import ListType, ModelType

from sofort.exceptions import RequestError, RequestErrors, SofortWarning
from sofort.internals import as_list

def response(xmlstr):
    result = xmltodict.parse(xmlstr)
    # only one root element is allowed in XML
    for root, value in result.iteritems():
        if value is None:
            return None
        factory = factories[root]
        return factory(value)


class ForcedListType(ListType):
    def to_native(self, value):
        return [self.field.to_native(item) for item in as_list(value)]


class SofortListType(ForcedListType):
    def __init__(self, field_name='', *args, **kwargs):
        ForcedListType.__init__(self, *args, **kwargs)
        self.field_name = field_name

    def to_native(self, value):
        return ForcedListType.to_native(self, value[self.field_name])


class Iso8601DateTimeType(DateTimeType):
    def to_native(self, value):
        return iso8601.parse_date(value)


class ErrorModel(Model):
    code = IntType()
    message = StringType()
    field = StringType()


class WarningModel(ErrorModel):
    def import_data(self, *args, **kwargs):
        ErrorModel.import_data(self, *args, **kwargs)
        warn(SofortWarning(**self.to_primitive()))


class SuErrorModel(Model):
    errors = SofortListType('error', ModelType(ErrorModel))


class RootErrorsModel(Model):
    error = ForcedListType(ModelType(ErrorModel))
    su = ModelType(SuErrorModel)


class NewTransactionModel(Model):
    transaction = StringType()
    payment_url = URLType()
    warnings = SofortListType('warning', ModelType(WarningModel))


class BankAccountModel(Model):
    holder = StringType()
    account_number = StringType()
    bank_code = StringType()
    bank_name = StringType()
    bic = StringType()
    iban = StringType()
    country_code = StringType()


class CostsModel(Model):
    fees = DecimalType()
    currency_code = StringType()
    exchange_rate = DecimalType()


class SuModel(Model):
    consumer_protection = BooleanType()


class StatusHistoryItemModel(Model):
    status = StringType()
    status_reason = StringType()
    time = Iso8601DateTimeType()


class TransactionDetailsModel(Model):
    project_id = IntType()
    transaction = StringType()
    test = BooleanType()
    time = Iso8601DateTimeType()
    status = StringType()
    status_reason = StringType()
    status_modified = Iso8601DateTimeType()
    payment_method = StringType()
    language_code = StringType()
    amount = DecimalType()
    amount_refunded = DecimalType()
    currency_code = StringType()
    reasons = SofortListType('reason', StringType())
    user_variables = SofortListType('user_variable', StringType())
    sender = ModelType(BankAccountModel)
    recipient = ModelType(BankAccountModel)
    email_customer = EmailType()
    phone_customer = StringType()
    exchange_rate = DecimalType()
    costs = ModelType(CostsModel)
    su = ModelType(SuModel)
    status_history_items = SofortListType('status_history_item',
                                          ModelType(StatusHistoryItemModel))


def transaction_list(transactions):
    return [TransactionDetailsModel(transact)
                for transact
                in as_list(transactions['transaction_details'])]


def error_handler(data):
    root = RootErrorsModel(data)
    errors = [RequestError(**error_item) for error_item in root.error]
    if root.su:
        errors.extend([RequestError(**error_item)
                            for error_item
                            in root.su.errors])
    raise RequestErrors(errors)


factories = {
    'errors': error_handler,
    'transactions': transaction_list,
    'new_transaction': NewTransactionModel,
}



import datetime
import requests

import sofort.xml

from sofort.exceptions import UnauthorizedError
from sofort.internals import Config, as_list
from sofort import model

from _version import __version__

API_URL = 'https://api.sofort.com/api/xml'
TRANSACTION_ID = '-TRANSACTION-'

TRANSACTION_HISTORY_LIMIT = datetime.timedelta(days=29)


class Client(object):
    """
    Sofort client. You can pass additional arguments to use them
    as default fall-back values when ``payment(...)`` request is
    performed.

    :param str user_id:
        User ID
    :param str api_key:
        API key
    :param str project_id:
        Project ID
    """
    def __init__(self, user_id, api_key, project_id, **kwargs):
        self.config = Config(
            base_url=API_URL,
            user_id=user_id,
            api_key=api_key,
            project_id=project_id,
            interface_version='python_sofort_v.{0}'.format(__version__),
            currency_code='EUR',
            country_code='DE',
            success_link_redirect=True,
        ).update(kwargs)

    def payment(self, amount, **kwargs):
        """Get payment URL and new transaction ID

        Usage::

            >>> import sofort
            >>> client = sofort.Client('123456', '123456', '123456',
                                       abort_url='https://mysite.com/abort')
            >>> t = client.pay(12, success_url='http://mysite.com?paid')

            >>> t.transaction
            123123-321231-56A3BE0E-ACAB
            >>> t.payment_url
            https://www.sofort.com/payment/go/136b2012718da216af4c20c2ec2f51100c90406e
        """
        params = self.config.clone()\
            .update({ 'amount': amount })\
            .update(kwargs)

        mandatory = ['abort_url', 'reasons', 'success_url']

        for field in mandatory:
            if not params.has(field):
                raise ValueError('Mandatory field "{}" is not specified'.format(field))

        params.reasons = [sofort.internals.strip_reason(reason)
                                for reason
                                in params.reasons]

        return self._request(sofort.xml.multipay(params), params)

    def details(self, transaction_ids):
        request_body = sofort.xml.transaction_request_by_params({
            'transaction': as_list(transaction_ids)
        })
        return self._request(request_body)

    def find_transactions(self, from_time=None, to_time=None, number=10,
                          **extra_params):
        today = datetime.datetime.now()

        if to_time is None:
            to_time = today

        if from_time is None:
            from_time = max(
                today - TRANSACTION_HISTORY_LIMIT,
                to_time - TRANSACTION_HISTORY_LIMIT
            )

        params = {
            'from_time': from_time,
            'to_time': to_time,
            'number': number
        }
        params.update(extra_params)
        request_body = sofort.xml.transaction_request_by_params(params)
        return self._request(request_body)

    def _request(self, data, config=None):
        if config is None:
            config = self.config.clone()

        response = self._request_xml(config, data).encode('utf-8')
        return model.response(response)

    def _request_xml(self, config, data):
        r = requests.post(config.base_url,
                          auth=(config.user_id, config.api_key),
                          data=data)

        if r.status_code == 200:
            return r.text
        elif r.status_code == 401:
            raise UnauthorizedError()
        elif r.status_code == 404:
            raise Exception('Sofort resource not found: {}'.format(
                                config.base_url))
        else:
            raise NotImplementedError(
                'Unknown response status: {}'.format(r.status_code)
            )



class UnauthorizedError(Exception):
    pass


class RequestErrors(Exception):
    def __init__(self, errors):
        super(RequestErrors, self).__init__('Request errors')
        self.errors = errors

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return '; '.join([str(error) for error in self.errors])


class RequestError(Exception):
    def __init__(self, code, message, field=''):
        self.code = code
        self.message = message
        self.field = field

    def __repr__(self):
        return self.code

    def __str__(self):
        result = str(self.message)
        if self.field is not None:
            result = '{}: {}'.format(self.field, result)
        return result


class SofortWarning(Warning):
    def __init__(self, code, message, field=''):
        self.code = code
        self.message = message
        self.field = field

    def __repr__(self):
        return "[{0.code}] {0.field}: {0.message}".format(self)

    def __str__(self):
        return repr(self)
